#!/bin/perl
#
# results.pl
#
# Gary Polhill 25 May 2001
#
# This is a script to analyse the results of experiments comparing two
# subpopulations against a third in one or more different environments. See the
# README file for more info.
#
# The script takes as argument a list of significance levels.

use strict;
use Cwd;
require "tools.pl";

my $tails = 1;
my $sameseed = 0;

# Check the arguments to the program

while($ARGV[0] =~ /^-/) {
  my $option = shift(@ARGV);
  if($option eq '-2tail') {
    $tails = 2;
  }
  elsif($option eq '-sameseed') {
    $sameseed = 1;
  }
  else {
    &usage("Invalid option: $option");
  }
}

my @siglevels = @ARGV;
if(scalar(@siglevels) == 0) {
  &usage("You must supply at least one significance level");
}

# Calculating the maximum significance level input
my $maxsiglevel = $ARGV[0];
my $numsiglevels = @ARGV;
my $numsigread = 1;
while ($numsigread < $numsiglevels) {
       if ($ARGV[$numsigread] > $maxsiglevel) {
           $maxsiglevel = $ARGV[$numsigread];
           }
       $numsigread++;
} 

# Get the list of environments being compared

my @envs = glob("*.model");
my @shortenvs = &get_short_envs(@envs);

# Create a log file to save the results in.

open(LOG, ">results.log") || die "Cannot create log file results.log: $!\n";

# Loop through the subpopulation directories and check the results

foreach my $spdir (glob("*-v-*-x-*")) {
  # Ignore non-directories that match the above glob
  next if !-d "$spdir";
  # Pull out the subpop labels from the directory name
  $spdir =~ /^(.*)-v-(.*)-x-(.*)$/;
  my $sp1 = $1;
  my $sp2 = $2;
  my $sp3 = $3;
  my $sp1summaryflag = 0;
  my $sp2summaryflag = 0;
  my $sp3summaryflag = 0;
  my $sp1summary;
  my $sp2summary;
  my $sp3summary;
  my $modelrun1 = "$sp1-v-$sp3";
  my $modelrun2 = "$sp2-v-$sp3";
  # Loop through all the environments
  for(my $env = 0; $env <= $#shortenvs; $env++) {
    my $envdir = "$shortenvs[$env]";
    # Check the environment contest sub-directory exists
    if(!-d "$spdir/$envdir") {
      die "Cannot find directory $spdir/$envdir -- have you run the ".
	"creation script (cwd = ".getcwd().")?\n";
    }
    # Change into the experiment contest sub-directory
    chdir "$spdir/$envdir" 
      || die "Cannot change directory to $spdir/$envdir: $!\n";
    # Declare a hash to use to store the seeds (all the runs in a particular
    # directory must have different seeds)
    my %seed;
    # Define variables to store the total number of times each subpopulation
    # did better when played against subpopulation 3.
    my($sp1sp3, $sp2sp3) = (0, 0, 0, 0);
    my($sp1eq, $sp2eq) = (0, 0);
    # Define a variable to store the total number of valid runs -- this is
    # where the subpopulations match, any paired bitstrings match, and the
    # seeds are all different
    my $validruns = 0;
    # Loop through all the runs
    my $run;
  RUN:
    for($run = 1;
	-f (my $reportfile1 = "$modelrun1-$shortenvs[$env]-run$run-rep.txt");
	$run++) {
      # Check there are report files for the runs using both environments
      my $reportfile2 = "$modelrun2-$shortenvs[$env]-run$run-rep.txt";
      if(!-f "$reportfile2") {
	die "Cannot find report file to pair with ".
	  "$spdir/$envdir/$reportfile1\n";
      }
      # Load in the report files
      my %report1;
      &load_report_file($reportfile1, \%report1);
      my %report2;
      &load_report_file($reportfile2, \%report2);
      my $msg;
      # Check paired replicates
      if($report1{'Parameter', 'Land Parcel file'} ne 'Not used') {
	my $ok = &compare_parcel_bitstrings(\%report1, \%report2, \$msg);
	if(!$ok) {
	  print LOG "WARNING: Run $run in directory $spdir/$envdir -- ".
	    " Paired land parcel match failed: $msg\n";
	  next;
	}
      }
      elsif($report2{'Parameter', 'Land Parcel file'} ne 'Not used') {
	print LOG "WARNING: Run $run in directory $spdir/$envdir -- ".
	  " Paired land parcel match failed: Model file not properly ",
	  "configured to do a paired run\n";
	next;
      }
      if($report1{'Parameter', 'Land Use file'} ne 'Not used') {
	my $ok = &compare_landuse_bitstrings(\%report1, \%report2);
	if(!$ok) {
	  print LOG "WARNING: Run $run in directory $spdir/$envdir -- ".
	    " Paired land use match failed\n";
	  next;
	}
      }
      elsif($report2{'Parameter', 'Land Use file'} ne 'Not used') {
	print LOG "WARNING: Run $run in directory $spdir/$envdir -- ".
	  " Paired land use match failed\n";
	next;
      }
      if($report1{'Parameter', 'Climate file'} ne 'Not used') {
	my $ok = &compare_climate_bitstrings(\%report1, \%report2);
	if(!$ok) {
	  print LOG "WARNING: Run $run in directory $spdir/$envdir -- ".
	    " Paired climate match failed\n";
	  next;
	}
      }
      elsif($report2{'Parameter', 'Climate file'} ne 'Not used') {
	print LOG "WARNING: Run $run in directory $spdir/$envdir -- ".
	  " Paired climate match failed\n";
	next;
      }
      if($report1{'Parameter', 'Economy file'} ne 'Not used') {
	my $ok = &compare_economy_bitstrings(\%report1, \%report2);
	if(!$ok) {
	  print LOG "WARNING: Run $run in directory $spdir/$envdir -- ".
	    " Paired economy match failed\n";
	  next;
	}
      }
      elsif($report2{'Parameter', 'Economy file'} ne 'Not used') {
	print LOG "WARNING: Run $run in directory $spdir/$envdir -- ".
	  " Paired economy match failed\n";
	next;
      }
      # Check seeds
      my $flag = 0;
      if($sameseed) {
	my $tmp1seed = $report1{'Parameter', 'Seed for simulation'};
	my $tmp2seed = $report2{'Parameter', 'Seed for simulation'};
	if($tmp1seed ne $tmp2seed) {
	  print LOG "WARNING: Report files $reportfile1 and $reportfile2 do ".
	    "not have the same seed ($tmp1seed and $tmp2seed respectively)\n";
	  $flag = 1;
	  $seed{$tmp2seed} = "$spdir/$envdir/$reportfile2";
	}
	if(defined($seed{$tmp1seed})) {
	  print LOG "WARNING: Seed $tmp1seed used in $seed{$tmp1seed} and ".
	    "$spdir/$envdir/$reportfile1\n";
	  $flag = 1;
	}
	$seed{$tmp1seed} = "$spdir/$envdir/$reportfile1";
      }
      else {
	foreach my $seedtmp ([$report1{'Parameter', 'Seed for simulation'},
			      $reportfile1],
			     [$report2{'Parameter', 'Seed for simulation'},
			      $reportfile2]) {
	  if(defined($seed{$$seedtmp[0]})) {
	    print LOG "WARNING: Seed $$seedtmp[0] used in $seed{$$seedtmp[0]}".
	      " and $spdir/$envdir/$$seedtmp[1]\n";
	    $flag = 1;
	    next;
	  }
	  $seed{$$seedtmp[0]} = "$spdir/$envdir/$$seedtmp[1]";
	}
      }
      next if $flag;
      # Check sub-populations -- all runs must use the same subpopulations
      # The create script makes sure that the subpopulation file is
      # consistent, so the subpopulations will always be in the same order.
      # The situation is complicated a little here because subpop1 only appears
      # in report1, subpop2 only in report2 and subpop3 in both reports, and
      # they all need to be checked.
      foreach my $subpop ([$sp1, \$sp1summaryflag, \$sp1summary,
			   [[\%report1, $reportfile1, 1],
			    # An id of 0 means it doesn't appear in the report
			    [\%report2, $reportfile2, 0]]],
			  [$sp2, \$sp2summaryflag, \$sp2summary,
			   # id of 0 means not in that report
			   [[\%report1, $reportfile1, 0],
			    [\%report2, $reportfile2, 1]]],
			  [$sp3, \$sp3summaryflag, \$sp3summary,
			   # sp3 appears in both reports as subpop 2
			   [[\%report1, $reportfile1, 2],
			    [\%report2, $reportfile2, 2]]]) {
	my($sp, $summaryflag, $summary, $reportids) = @$subpop;
	foreach my $reportid (@$reportids) {
	  my($report, $reportfile, $id) = @$reportid;
	  # Not in the report if id 0, so move on to the next report
	  next if $id == 0;
	  if(!$$summaryflag) {
	    $$summaryflag = 1;
	    $$summary = $$report{'SubpopSummary', $id};
	    print LOG "NOTE: $sp is $$summary in $spdir/*\n";
	  }
	  elsif($$summary ne $$report{'SubpopSummary', $id}) {
	    print LOG "WARNING: Inconsistencies for $sp -- not the same in ".
	      "previous runs as in $spdir/$envdir/$reportfile\n\tPreviously:".
		"\t$$summary\n\tCurrently:\t$$report{'SubpopSummary', $id}\n";
	    next RUN;
	  }
	}
      }
      
      # Find out the year of the report and check they are the same
      my $year = pop(@{$report1{'ParcelSubPopReport', 'ReportingYears'}});
      if($year != pop(@{$report2{'ParcelSubPopReport', 'ReportingYears'}})) {
	print LOG "WARNING: $modelrun1 and $modelrun2 report".
	  " on different final years for ParcelSubPopReport in".
	    " $spdir/$envdir, run $run\n";
	next;
      }
      # OK, so now, who did better than whom?
      # Have to use temporary variables to store the associative arrays
      # dereferenced from the report for programmer sanity -- XEmacs can't
      # cope with the syntax for indenting purposes (even though there's
      # no ${!!)
      my %r1sp = %{$report1{'ParcelSubPopReport', $year}};
      my %r2sp = %{$report2{'ParcelSubPopReport', $year}};
      # So now use $r1sp{1} instead of
      # ${$report1{'ParcelSubPopReport', $year}}{1}
      # subpop1 did better than subpop2 against subpop3
      $sp1sp3++ if($r1sp{1} > $r2sp{1});
      # subpop2 did better than subpop1 against subpop3
      $sp2sp3++ if($r2sp{1} > $r1sp{1});
      # there was no difference in performance
      $sp1eq++ if($r1sp{1} == $r2sp{1});
      $sp2eq++ if($r1sp{1} == $r2sp{1});
      # Increment the number of valid runs
      $validruns++;
    }
    if(-f (my $reportfile2 = "$modelrun2-$shortenvs[$env]-run$run-rep.txt")) {
      die "Cannot find report file to pair with ".
	"$spdir/$envdir/$reportfile2\n";
    }
    # Print the p-value for the results in this run using a 1-tailed
    # binomial test 
    if($validruns > 0) {
      if($tails == 1) {
        open(RESULTS, ">results.db")
                            || die "Cannot create results.db $!\n";
	print "Results for $spdir/$envdir (1-tailed sign test of ".
	  "$validruns valid runs)\n";
	print "Hypothesis\tSignificance\tp-value\tWins\tNon-tied runs\n";
	print RESULTS "Results for $spdir/$envdir (1-tailed sign test of ".
	  "$validruns valid runs)\n";
        print RESULTS "Test type:\t2\n";
        print RESULTS "Tails:\t1\n";
        print RESULTS "Maximum significance level:\t$maxsiglevel\n";
	foreach my $test ([$sp1, $sp2, $sp3, $sp1sp3, $sp1eq],
			  [$sp2, $sp1, $sp3, $sp2sp3, $sp2eq]) {
	  my($sp, $vsp, $xsp, $score, $nodiff) = @$test;
	  my $pvalue;
	  # Find the smallest of the command-line entered p-values that
	  # is greater than this p-value.
	  my $significance = "NS";
	  my $minlevel = 1;
	  foreach my $userlevel (@siglevels) {
	    my $significant = &testgtbinomial($validruns - $nodiff, $score,
					      $userlevel, \$pvalue);
	    if($significant && $userlevel < $minlevel) {
	      $minlevel = $userlevel;
	      $significance = $userlevel;
	    }
	  }
	  print "$sp does better against $xsp than $vsp in environment ".
	    "$shortenvs[$env]\t$significance\t$pvalue\t$score\t",
	    $validruns - $nodiff, "\n";
          print RESULTS "Hyp/Res/p_value:";
	  print RESULTS "\t$sp does better against $xsp than $vsp in ".
                     "environment $shortenvs[$env]\t$significance\t$pvalue\n";
	}
        close(RESULTS);
      }
      elsif($tails == 2) {
        open(RESULTS, ">results.db")
                            || die "Cannot create results.db $!\n";
	print "Results for $spdir/$envdir (2-tailed sign test of ".
	  "$validruns valid runs)\n";
	print "Hypothesis\tSignificance\tp-value\t$sp1 wins\t$sp2 wins\t".
	  "Non-tied runs\n";
	print RESULTS "Results for $spdir/$envdir (2-tailed sign test of ".
	  "$validruns valid runs)\n";
        print RESULTS "Test type:\t2\n";
        print RESULTS "Tails:\t2\n";
        print RESULTS "Maximum significance level:\t$maxsiglevel\n";
	my $pvalue;
	# Find the smallest of the command-line entered p-values that
	# is greater than this p-value.
	my $significance = "NS";
	my $minlevel = 1;
	foreach my $userlevel (@siglevels) {
	  my $significant = &testnebinomial($validruns - $sp1eq, $sp1sp3,
					    $userlevel, \$pvalue);
	  if($significant && $userlevel < $minlevel) {
	    $minlevel = $userlevel;
	    $significance = $userlevel;
	  }
	}
        print RESULTS "Hyp/Res/p_value:\t";
	if($sp1sp3 > $sp2sp3) {
	  print "$sp1 and $sp2 are not equal against $sp3 in environment ".
	    "$shortenvs[$env]";
    print RESULTS "$sp1 and $sp2 are not equal against $sp3 in environment ".
	    "$shortenvs[$env]";
	}
	else {
	  print "$sp2 and $sp1 are not equal against $sp3 in environment ".
	    "$shortenvs[$env]";
    print RESULTS "$sp2 and $sp1 are not equal against $sp3 in environment ".
	    "$shortenvs[$env]";
	}
	print "\t$significance\t$pvalue\t$sp1sp3\t$sp2sp3\t",
	$validruns - $sp1eq, "\n";
	print RESULTS "\t$significance\t$pvalue\n";
      }
        close(RESULTS);
    }
    else {
      print "There were no valid runs in $spdir/$envdir\n";
    }
    # Change directory back to cwd ready for next set of runs
    chdir "$ENV{PWD}";
  }
}
close(LOG);
print "Results completed. Check results.log for any messages\n";
exit 0;

sub usage {
  my($message) = @_;
  die <<DEATH;
$message

This script is called as follows:
$0 [-2tail] [-sameseed] <list of significance levels>

This program looks in the report files of the runs completed and does a
paired-replicate sign test at the significance levels specified to see if one
sub-population does better than another sub-population when played against
a third sub-population.

The -2tail option is specified to do a two-tailed binomial test instead of one.

The -sameseed option is given if the same option was applied to the run script.
This affects the process of checking for seeds. Paired runs are invalid if they
do not use the same seed, or if a seed has been used in another paired run. The
default is to use different seeds for each run (including each of a pair), 
creating paired runs through the use of files to load in the biophysical
properties, land uses, climate and economy, and complain if any run uses a seed
that has been used by any other run.
DEATH
}
