#!/bin/perl
#
# run.pl
#
# Gary Polhill 24 May 2001
#
# This is a script to do paired replicate runs for 2 subpopulations comparing
# their performance against a third. It expects the corresponding create.pl to
# have been run already in the current working directory.
#
# The program's usage is as follows:
#
# run.pl [-history[=<number of history runs to save>]] [-restart]
#        [-command=<path to FEARLUS command>]
#        [-postinitseed[=<integer value of post initiation seed>]]
#        [-onlydirs=<list of sp directories to run>]
#        <number of runs to do>

use strict;
use Cwd;
require "tools.pl";

srand(time);
# Global variable declarations

# $history is -1 for all runs, 0 for no runs, +ve for first n runs
my $history = 0;

# It is possible to control the seed used post initialisation
# A flag variable is used to hold whether or not the argument was used
my $postinitseed = 0;
my $postinitseedpres = 0;

# Allow the user to have all warnings saved to run.log
my $savewarns = 0;

# Use the same seed for each pair of runs, instead of sharing parameter files.
# Note that this option needs to be used with caution, as each pair of runs
# must involve the same number of calls to the random number generator during
# initialisation.
my $sameseed = 0;

# $restart is 0 to continue with runs so far and 1 to start from the beginning
my $restart = 0;

# $exe is the command to use to run the FEARLUS model
my $exe = '/home/gp/swarm/fearlus/model0-6-6/model0-6-6';

# Number of runs
my $runs;

# SP directories to do
my @onlyspdirs;
my %runspdirs;
my $allspdirs = 1;

# Check the arguments to the program

my $arg;

$arg = shift(@ARGV) || &usage("You must supply the number of runs");
while($arg =~ /^-/) {
  if($arg =~ /^-history/) {
    if($arg =~ /^-history=([0-9]+)$/) {
      $history = $1;
    }
    elsif($arg =~ /^-history.+/) {
      &usage("Garbled history argument: $arg");
    }
    else {
      $history = -1;
    }
  }
  elsif($arg =~ /^-postinitseed/) {
    if($arg =~ /^-postinitseed=([0-9]+)$/) {
      $postinitseed = $1;
      $postinitseedpres = 1;
    }
    else {
      &usage("Garbled postinitseed argument: $arg");
    }
  }
  elsif($arg =~ /^-onlydirs=(.+)$/) {
    my $dirlist = $1;
    @onlyspdirs = split(/,/, $dirlist);
    $allspdirs = 0;
    $restart = 1;
    foreach my $dir (@onlyspdirs) {
      if(!-d "$dir") {
	die "$dir is not a valid sub-population directory\n";
      }
      $runspdirs{$dir} = $dir;
    }
  }
  elsif($arg eq '-restart') {
    $restart = 1;
  }
  elsif($arg eq '-sameseed') {
    $sameseed = 1;
    warn "WARNING: Sameseed option means that paired runs will not ".
      "have the same initialisation unless each involves the same calls to ".
	"the RNG at the same times.\n";
  }
  elsif($arg eq '-savewarns') {
    $savewarns = 1;
  }
  elsif($arg =~ /^-command=(.+)$/) {
    $exe = $1;
  }
  $arg = shift(@ARGV) || &usage("You must supply the number of runs");
}

if($arg =~ /^[0-9]+$/) {
  $runs = $arg;
}
else {
  &usage("Invalid number of runs: $arg");
}

# Check the command is runnable

if(!-x "$exe") {
  die "$exe is not an existing file that you have execute permissions for".
    ", so I cannot run the model\n";
}
if($exe !~ /\//) {
  die "$exe must be an absolute path to the FEARLUS model executable\n";
}

# Get the list of environments we are comparing

my @envs = glob("*.model");
my @shortenvs = &get_short_envs(@envs);

# Create a log file to save messages in. If the log file exists, check progress
# through the experiment. Log file entries indicating successful completion
# of a run have the format <DATE> Completed run <RUN> of environment <ENV> in
# directory <DIR>
#
# The presence of a log file will be used to allow continuation of interrupted
# experiments

# $oktogo is a flag that tells the loop it is ready to run. It will only be 0
# at the beginning of the loop if there are a number of runs that have been
# done already.
my $oktogo = 1;
my($godate, $gorun, $gocontest, $godir);
if(-f "run.log" && !$restart) {
  open(LOG, "<run.log") || die "Cannot read log file run.log: $!\n";
  while(my $line = <LOG>) {
    if($line 
       =~ m/^(.+) Completed run (.+) of contest (.+) in directory (.+)$/) {
      # We just want the last completed run that appears in the file
      ($godate, $gorun, $gocontest, $godir) = ($1, $2, $3, $4);
      $oktogo = 0;
    }
  }
  close(LOG);
  open(LOG, ">>run.log") || die "Cannot append to log file run.log: $!\n";
  print LOG &datetime()." Restarting after run $gorun of contest $gocontest".
    " in directory $godir\n";
}
elsif(-f "run.log") {
  open(LOG, ">>run.log") || die "Cannot append to log file run.log: $!\n";
}
else {
  open(LOG, ">run.log") || die "Cannot create log file run.log: $!\n";
}

if(!$allspdirs) {
  print LOG &datetime()." Running experiments in the following subpopulation ".
    "directories only: ".join(", ", sort(keys(%runspdirs)))."\n";
}

my $oldfh = select LOG;
# Set autoflush on LOG to 1 -- this will flush the buffer after every print.
$| = 1;
select $oldfh;

# Loop through the directories and run the experiments therein

foreach my $spdir (sort(glob("*-v-*-x-*"))) {
  # The create script skips over matching files that aren't directories
  next if !-d "$spdir";
  next if !($allspdirs || defined($runspdirs{$spdir}));
  $spdir =~ /^(.*)-v-(.*)-x-(.*)$/;
  my $sp1 = $1;
  my $sp2 = $2;
  my $sp3 = $3;
  # Loop through all the combinations of the environments
  for(my $env = 0; $env <= $#shortenvs; $env++) {
    my $envdir = "$shortenvs[$env]";
    # Check the environment contest sub-directory exists
    if(!-d "$spdir/$envdir") {
      die "Cannot find directory $spdir/$envdir -- have you run the ".
	"creation script (cwd = ".getcwd().")?\n";
    }
    # Loop again if we're not OK to go
    next if (!$oktogo && $godir ne "$spdir/$envdir");
    # Change into the experiment contest sub-directory
    chdir "$spdir/$envdir" 
      || die "Cannot change directory to $spdir/$envdir: $!\n";
    # Loop through all the runs
    for(my $run = 1; $run <= $runs; $run++) {
      my $seedpair;		# Used for -sameseed option
      my $firstpair = 1;	# Used for -sameseed option
      next if (!$oktogo && $gorun != $run);
      # Loop through each of the pair of model files we've got to do
      foreach my $modelrun ("$sp1-v-$sp3", "$sp2-v-$sp3") {
	my $modelfile = "$modelrun-$envs[$env]";
	next if (!$oktogo && $gocontest ne $modelrun);
	# If we get here and we're not $oktogo, then we're now at the point
	# of the last successful run. Set $oktogo to 1 and loop again.
	if(!$oktogo && $gocontest eq $modelrun) {
	  print LOG &datetime()." Reached run $run of contest $modelrun in ".
	    "directory $spdir/$envdir\n";
	  $oktogo = 1;
	  next;
	}
	# If we get here, then we're $oktogo. Call a subroutine to create
	# a new modelfile (if necessary) for the run.
	my $newmodelfile = &create_run_modelfile($modelfile, $run);
	if($sameseed) {
	  &no_use_bitstring_files($newmodelfile);
				# This removes the useXXXXFile flag settings
				# from the model file
	}
	# Set up the other commandline argument filenames
	my $reportfile = "$modelrun-$shortenvs[$env]-run$run-rep";
	my $historyfile = "$modelrun-$shortenvs[$env]-run$run";
	# Delete any files that the report or history files might over-write
	if(-e "$reportfile.txt") {
	  unlink "$reportfile.txt"
	    || die "Cannot remove $reportfile.txt in directory ".
	      "$spdir/$envdir: $!\n";
	}
	# Delete the history file regardless of whether it exists from an
	# earlier run -- it won't correspond to this run, and that could be
	# confusing
	if(-e "$historyfile.txt") {
	  unlink "$historyfile.txt"
	    || die "Cannot remove $historyfile.txt in directory ".
	      "$spdir/$envdir: $!\n";
	}
	# Create a temporary file to store any error messages
	my $tmpfile;
	do {
	  $tmpfile = sprintf("/tmp/FEARLUS%06d", int(rand() * 999999));
	} while(-e "$tmpfile");
	my $tmpoutfile = "/dev/null";
	my $seedtouse = "TIME";

	# Set up the command to run
	my $command;
	if($sameseed) {
	  if($firstpair) {
	    do {
	      $tmpoutfile = sprintf("/tmp/FEARLUSP%06d", int(rand() * 999999));
	    } while(-e "$tmpoutfile");
	  }
	  else {
	    $seedtouse = $seedpair;
	  }
	}
	$command = "$exe --batch --withseed=$seedtouse ".
	  "--repconfig=Parcel.repcfg --report=$reportfile ".
	    "--parameters=$newmodelfile ".
	      (($postinitseedpres == 1) ?
	       "--postinitseed=$postinitseed " : "").
		 (($history < 0) || ($history >= $run) ?
		  "--history=$historyfile " : "").
		    "> $tmpoutfile 2> $tmpfile";
	print LOG &datetime()." Running command $command\n";
	# Run the command
	my $retval = system("$command");
	if($retval != 0) {
	  print LOG &datetime()." Command failed with exit code $retval:\n";
	}
	elsif($savewarns) {
	  print LOG &datetime()." Warning messages (if any):\n";
	}
	if($retval !=0 || $savewarns) {
	  open(TMP, "<$tmpfile") 
	    || die "Cannot read tempfile $tmpfile: $!\n";
	  while(<TMP>) {
	    print LOG &datetime()."   $_";
	  }
	  close(TMP);
	}
	if($retval == 0) {
	  print LOG &datetime()." Completed run $run of contest ".
	    "$modelrun in directory $spdir/$envdir\n";
	}
	unlink($tmpfile) || die "Cannot delete tempfile $tmpfile: $!\n";

	# Set things up for -sameseed option

	if($sameseed && $firstpair) {
	  open(TMPOUT, "<$tmpoutfile") 
	    || die "Cannot read tempoutfile $tmpoutfile: $!\n";
	  my $seedfound = 0;
	  while(my $line = <TMPOUT>) {
	    $line = &mychomp($line);
	    if($line =~ /^Using seed: ([0-9]+)$/) {
	      $seedpair = $1;
	      $seedfound = 1;
	      last;
	    }
	  }
	  close(TMPOUT);
	  die "Cannot find seed in tempoutfile $tmpoutfile\n" if !$seedfound;
	  unlink($tmpoutfile)
	    || die "Cannot delete tempoutfile $tmpoutfile: $!\n";
	}
	$firstpair = 0;		# Used for -sameseed option
      }
    }
    chdir "$ENV{PWD}" 
      || die "Cannot change directory back to $ENV{PWD}: $!\n";
  }
}

exit 0;

sub usage {
  my($message) = @_;
  die <<DEATH;
$message

This script is called as follows:
$0 [-history[=<number of history runs to save>]] [-restart] [-savewarns] 
   [-postinitseed=<seed to be used post initialisation>] [-sameseed]
   [-command=<full path to FEARLUS model executable>] <number of runs to do>


-history is an optional argument that will tell FEARLUS to save a history file
for each run that it does. You can optionally specify that the first NN runs
save a history (whilst the rest do not) by giving -history=NN. Be warned that
history files can occupy a lot of disk space, especially when making a number
of runs.

-restart will cause the script to restart the experiment even if it is part of
the way through it. If you do not give -restart, then the script will check
run.log for successfully completed runs, and not do them again.

-onlydirs allows you to specify a comma separated (no spaces) list of subpop
directories in which to run experiments. Experiments in these directories will
be run regardless of what was done before.

-postinitseed this entered along with an integer allows you to specify the
seed to be used post initialisation.

-saveseed allows you to perform paired runs using the same seed for each run.
Only use this option if you are certain that initialisation for each run of a
pair will involve the same calls to the random number generator at the same
times.

-command specifies an alternative path to the FEARLUS model. The default is
currently $exe.

-savewarns copies all output to stderr from successful runs into run.log. The
default is only to do this when the run is unsuccessful.


The final argument is the number of runs to do in each experiment.
DEATH
}
